@@ -1,3 +1,7 @@
+/*
+    Copyright Michael Lodder. All Rights Reserved.
+    SPDX-License-Identifier: Apache-2.0
+*/
 use crate::{get_mod, GcdResult};
 use gmp::{
     mpz::{Mpz, ProbabPrimeResult},
@@ -224,6 +228,9 @@ impl Bn {
     /// Generate a safe prime with `size` bits
     pub fn safe_prime(size: usize) -> Self {
         let mut rand_state = RandState::new();
+        let mut seed = [0u8; 32];
+        rand::rngs::OsRng::default().fill_bytes(&mut seed);
+        rand_state.seed(Mpz::from(&seed[..]));
         let mut p = (rand_state.urandom_2exp((size - 1) as u64).nextprime() << 1) + 1;
         loop {
             while p.bit_length() != size {
@@ -239,6 +246,9 @@ impl Bn {
     /// Generate a prime with `size` bits
     pub fn prime(size: usize) -> Self {
         let mut rand_state = RandState::new();
+        let mut seed = [0u8; 32];
+        rand::rngs::OsRng::default().fill_bytes(&mut seed);
+        rand_state.seed(Mpz::from(&seed[..]));
         let mut p = rand_state.urandom_2exp(size as u64).nextprime();
         while p.bit_length() != size {
             p = rand_state.urandom_2exp(size as u64).nextprime();
@@ -253,6 +263,70 @@ impl Bn {
             _ => false,
         }
     }
+
+    /// Simultaneous integer division and modulus
+    pub fn div_rem(&self, other: &Self) -> (Self, Self) {
+        let (q, r) = _div_rem(&self.0, &other.0);
+        (Self(q), Self(r))
+    }
+}
+
+#[repr(C)]
+pub struct MpzStruct {
+    _mp_alloc: std::os::raw::c_int,
+    _mp_size: std::os::raw::c_int,
+    _mp_d: *mut std::os::raw::c_void,
+}
+pub type MpzSrcptr = *const MpzStruct;
+pub type MpzPtr = *mut MpzStruct;
+
+#[link(name = "gmp")]
+extern "C" {
+    fn __gmpz_init(x: MpzPtr);
+    fn __gmpz_tdiv_qr(q: MpzPtr, r: MpzPtr, n: MpzSrcptr, d: MpzSrcptr);
+    fn __gmpz_clear(x: MpzPtr);
+}
+
+pub struct MpzMpz {
+    mpz: MpzStruct,
+}
+
+unsafe impl Send for MpzMpz {}
+unsafe impl Sync for MpzMpz {}
+
+impl Drop for MpzMpz {
+    fn drop(&mut self) {
+        unsafe { __gmpz_clear(&mut self.mpz) }
+    }
+}
+
+impl Default for MpzMpz {
+    fn default() -> MpzMpz {
+        unsafe {
+            let mut mpz = std::mem::MaybeUninit::uninit().assume_init();
+            __gmpz_init(&mut mpz);
+            MpzMpz { mpz }
+        }
+    }
+}
+
+fn _div_rem(lhs: &Mpz, rhs: &Mpz) -> (Mpz, Mpz) {
+    unsafe {
+        if rhs.is_zero() {
+            panic!("divide by zero");
+        }
+
+        let tv1: MpzMpz = std::mem::transmute(lhs.clone());
+        let tv2: MpzMpz = std::mem::transmute(rhs.clone());
+
+        let mut q_res = MpzMpz::default();
+        let mut r_res = MpzMpz::default();
+        __gmpz_tdiv_qr(&mut q_res.mpz, &mut r_res.mpz, &tv1.mpz, &tv2.mpz);
+
+        let q = std::mem::transmute(q_res);
+        let r = std::mem::transmute(r_res);
+        (q, r)
+    }
 }
 
 #[test]
@@ -260,6 +334,28 @@ fn safe_prime() {
     let n = Bn::safe_prime(1024);
     assert_eq!(n.0.bit_length(), 1024);
     assert!(n.is_prime());
-    let sg: Bn = n >> 1;
-    assert!(sg.is_prime())
+    let sg: Bn = &n >> 1;
+    assert!(sg.is_prime());
+    // Make sure it doesn't produce the same prime when called twice
+    let m = Bn::safe_prime(1024);
+    assert_eq!(m.0.bit_length(), 1024);
+    assert!(m.is_prime());
+    let sg: Bn = &m >> 1;
+    assert!(sg.is_prime());
+    assert_ne!(n, m);
+}
+
+#[test]
+fn div_rem_test() {
+    let a = Bn::from(11);
+    let b = Bn::from(3);
+    let (q, r) = a.div_rem(&b);
+    assert_eq!(q, Bn::from(3));
+    assert_eq!(r, Bn::from(2));
+
+    let a = Bn::from(23);
+    let b = Bn::from(10);
+    let (q, r) = a.div_rem(&b);
+    assert_eq!(q, Bn::from(2));
+    assert_eq!(r, Bn::from(3));
 }
